import sys
import textwrap
import uuid
from functools import partial
from typing import Union

import eventlet
from cpl_core.dependency_injection import ServiceProviderABC
from cpl_core.utils import CredentialManager
from eventlet import wsgi
from flask import Flask, request, jsonify, Response
from flask_cors import CORS
from flask_socketio import SocketIO
from werkzeug.exceptions import NotFound

from bot_api.configuration.api_settings import ApiSettings
from bot_api.configuration.authentication_settings import AuthenticationSettings
from bot_api.exception.service_error_code_enum import ServiceErrorCode
from bot_api.exception.service_exception import ServiceException
from bot_api.logging.api_logger import ApiLogger
from bot_api.model.error_dto import ErrorDTO
from bot_api.route.route import Route


class Api(Flask):
    def __init__(
        self,
        logger: ApiLogger,
        services: ServiceProviderABC,
        api_settings: ApiSettings,
        auth_settings: AuthenticationSettings,
        *args,
        **kwargs,
    ):
        if not args:
            kwargs.setdefault("import_name", __name__)

        Flask.__init__(self, *args, **kwargs)

        self._logger = logger
        self._services = services
        self._api_settings = api_settings
        self._auth_settings = auth_settings

        self._cors = CORS(self, support_credentials=True)

        # register hooks
        self.before_request(self.before_request_hook)
        self.after_request(self.after_request_hook)

        # register error handler
        exc_class, code = self._get_exc_class_and_code(Exception)
        self.register_error_handler(exc_class, self.handle_exception)

        # websockets
        self._socketio = SocketIO(self, cors_allowed_origins="*", path="/api/socket.io")
        self._socketio.on_event("connect", self.on_connect)
        self._socketio.on_event("disconnect", self.on_disconnect)

        self._requests = {}

    @staticmethod
    def _get_methods_from_registered_route() -> Union[list[str], str]:
        methods = ["Unknown"]
        if (
            request.path in Route.registered_routes
            and len(Route.registered_routes[request.path]) >= 1
            and "methods" in Route.registered_routes[request.path][1]
        ):
            methods = Route.registered_routes[request.path][1]["methods"]

        if len(methods) == 1:
            return methods[0]
        return methods

    def _register_routes(self):
        for path, f in Route.registered_routes.items():
            route = f[0]
            kwargs = f[1]
            cls = None
            qual_name_split = route.__qualname__.split(".")
            if len(qual_name_split) > 0:
                cls_type = vars(sys.modules[route.__module__])[qual_name_split[0]]
                cls = self._services.get_service(cls_type)

            partial_f = partial(route, self if cls is None else cls)
            partial_f.__name__ = route.__name__
            self.route(path, **kwargs)(partial_f)

    def handle_exception(self, e: Exception):
        self._logger.error(__name__, f"Caught error", e)

        if isinstance(e, ServiceException):
            ex: ServiceException = e
            self._logger.error(__name__, ex.get_detailed_message())
            error = ErrorDTO(ex.error_code, ex.message)
            return jsonify(error.to_dict()), 500
        elif isinstance(e, NotFound):
            self._logger.error(__name__, e.description)
            error = ErrorDTO(ServiceErrorCode.NotFound, e.description)
            return jsonify(error.to_dict()), 404
        else:
            tracking_id = uuid.uuid4()
            user_message = f"Tracking Id: {tracking_id}"
            self._logger.error(__name__, user_message, e)
            error = ErrorDTO(None, user_message)
            return jsonify(error.to_dict()), 400

    def before_request_hook(self):
        request_id = uuid.uuid4()
        self._requests[request] = request_id
        method = request.access_control_request_method

        self._logger.info(
            __name__,
            f"Received {request_id} @ {self._get_methods_from_registered_route() if method is None else method} {request.url} from {request.remote_addr}",
        )

        headers = str(request.headers).replace("\n", "\n\t\t")
        data = request.get_data()
        data = "" if len(data) == 0 else str(data.decode(encoding="utf-8"))

        text = textwrap.dedent(
            f"Request: {request_id}:\n\tHeader:\n\t\t{headers}\n\tUser-Agent: {request.user_agent.string}\n\tBody: {data}"
        )
        self._logger.trace(__name__, text)

    def after_request_hook(self, response: Response):
        method = request.access_control_request_method
        request_id = f"{self._get_methods_from_registered_route() if method is None else method} {request.url} from {request.remote_addr}"
        if request in self._requests:
            request_id = self._requests[request]

        self._logger.info(__name__, f"Answered {request_id}")

        headers = str(request.headers).replace("\n", "\n\t\t")
        data = request.get_data()
        data = "" if len(data) == 0 else str(data.decode(encoding="utf-8"))

        text = textwrap.dedent(f"Request: {request_id}:\n\tHeader:\n\t\t{headers}\n\tResponse: {data}")
        self._logger.trace(__name__, text)

        return response

    def start(self):
        self._logger.info(__name__, f"Starting API {self._api_settings.host}:{self._api_settings.port}")
        self._register_routes()
        self.secret_key = CredentialManager.decrypt(self._auth_settings.secret_key)
        # from waitress import serve
        # https://docs.pylonsproject.org/projects/waitress/en/stable/arguments.html
        # serve(self, host=self._apt_settings.host, port=self._apt_settings.port, threads=10, connection_limit=1000, channel_timeout=10)
        wsgi.server(eventlet.listen((self._api_settings.host, self._api_settings.port)), self, log_output=False)

    def on_connect(self):
        self._logger.info(__name__, f"Client connected")

    def on_disconnect(self):
        self._logger.info(__name__, f"Client disconnected")
