from datetime import datetime

from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_discord.service import DiscordBotServiceABC

from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_joined_game_server_repository_abc import UserJoinedGameServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.user_joined_game_server import UserJoinedGameServer
from bot_data.model.user_role_enum import UserRoleEnum
from bot_graphql.abc.query_abc import QueryABC
from modules.base.abc.base_helper_abc import BaseHelperABC
from modules.base.configuration.base_server_settings import BaseServerSettings
from modules.permission.service.permission_service import PermissionService


class UserJoinedGameServerMutation(QueryABC):
    def __init__(
        self,
        logger: LoggerABC,
        base_helper: BaseHelperABC,
        servers: ServerRepositoryABC,
        users: UserRepositoryABC,
        user_joined_game_servers: UserJoinedGameServerRepositoryABC,
        bot: DiscordBotServiceABC,
        db: DatabaseContextABC,
        permissions: PermissionService,
    ):
        QueryABC.__init__(self, "UserJoinedGameServerMutation")

        self._logger = logger
        self._base_helper = base_helper
        self._servers = servers
        self._users = users
        self._user_joined_game_servers = user_joined_game_servers
        self._bot = bot
        self._db = db
        self._permissions = permissions

        self.set_field("userJoined", self.resolve_user_joined)
        self.set_field("userLeaved", self.resolve_user_leaved)

    def resolve_user_joined(self, *_, input: dict):
        user = self._users.get_user_by_id(input["userId"])
        self._can_user_mutate_data(user.server, UserRoleEnum.admin)

        active = self._user_joined_game_servers.find_active_user_joined_game_server_by_user_id(user.id)
        if active is not None:
            self._logger.debug(
                __name__,
                f"Skip UserJoinedGameServer for user {user.id}. User already plays on {active.game_server}.",
            )
            return

        new = UserJoinedGameServer(user, input["gameServer"], datetime.now())
        self._user_joined_game_servers.add_user_joined_game_server(new)
        self._db.save_changes()

        return self._user_joined_game_servers.get_active_user_joined_game_server_by_user_id(user.id)

    def resolve_user_leaved(self, *_, input: dict):
        user = self._users.get_user_by_id(input["userId"])
        self._can_user_mutate_data(user.server, UserRoleEnum.admin)

        active = self._user_joined_game_servers.find_active_user_joined_game_server_by_user_id(user.id)
        if active is None:
            return None
        active.leaved_on = datetime.now()

        settings: BaseServerSettings = self._base_helper.get_config(user.server.discord_id)

        ontime = round((active.leaved_on - active.joined_on).total_seconds() / 3600, 2)
        old_xp = user.xp
        user.xp += round(ontime * settings.xp_per_ontime_hour)

        self._user_joined_game_servers.update_user_joined_game_server(active)
        self._db.save_changes()
        self._logger.debug(
            __name__,
            f"User {user} leaved_on {active.leaved_on}. Ontime: {ontime}h | xp: from {old_xp} to {user.xp}",
        )

        return active
